local addon = CreateFrame("Frame")
_G.Broker_Imba = addon

-----------------------------------------------------------------------
-- Raw data
--

local abilityMap = {
	["+5% Crit"] = 1,
	["+10% AP"] = 2,
	["+6% SP"] = 4,
	["+10% SP"] = 8,
	["+10% Phys. Haste"] = 16,
	["+5% Spell Haste"] = 32,
	["+5% Stats"] = 64,
	["+3% Dmg"] = 128,
	["+Str/Agi"] = 256,
	["+Sta"] = 512,
	["BL/Hero"] = 1024,
	["+Int"] = 2048,
	["+Manareg"] = 4096,
	["Replenishment"] = 8192,

	["+5% Spell Crit."] = 16384,
	["+4% Phys. Dmg"] = 32768,
	["+8% Spell Dmg"] = 65536,
	["+30% Bleed Dmg"] = 131072,
	["-Attack Speed"] = 262144,
	["-Cast Speed"] = 524288,
	["-Phys. Dmg"] = 1048576,
	["-12% Armor"] = 2097152,
	["-25% Healing"] = 4194304,
}
local buffs, debuffs = {}, {}
for k, v in pairs(abilityMap) do
	if v < 8193 then buffs[k] = true
	else debuffs[k] = true end
end

local totalBits = {
	HUNTER = { 4194304, 4194432, 4194306, 4194320 },
	WARRIOR = { 262912, 4621056, 4457217, 3408640 },
	SHAMAN = { 5428, 5437, 5430, 5428 },
	MAGE = { 3076, 527492, 19468, 11268 },
	PRIEST = { 512, 512, 512, 4203040 },
	WARLOCK = { 589824, 589824, 589832, 614400 },
	ROGUE = { 6815744, 6881280, 6848512, 6946817 },
	DRUID = { 64, 65632, 3539009, 8256 },
	PALADIN = { 4162, 4162, 1314882, 12482 },
	DEATHKNIGHT = { 262402, 1310978, 295186, 327938 },
	Devilsaur = 4194305,
	Fox = 262144,
	Worm = 32768,
	Cat = 256,
	Ravager = 32768,
	Wolf = 1,
	Serpent = 2097152,
	["Shale Spider"] = 64,
	Bear = 1048576,
	Tallstrider = 262144,
	Sporebat = 524288,
	["Core Hound"] = 525312,
	Dragonhawk = 65536,
	Boar = 131072,
	Silithid = 512,
	["Wind Serpent"] = 65536,
	Rhino = 131072,
	Raptor = 2097152,
	Felhunter = 6144,
	["Spirit Beast"] = 256,
	Hyena = 131072,
}
local L = GetLocale()
if L == "deDE" then
	totalBits.GermanWordForFox = totalBits.Fox
	-- etc...
end

addon.totalBits = totalBits -- debug

-----------------------------------------------------------------------
-- Local variables
--

-- Holds a map of group members and bitwise what they provide
local providers = {}
-- Number of abilities that the current group covers (max 23).
local abilitiesCovered = 0
-- Missing abilites (for fast tooltip access)
local missingBuffs = {}
local missingDebuffs = {}
-- List of members that need a scan
local needsScanning = {}
-- Units that we need to watch and update later when we can
local watch = {}
-- Last time we scanned
local lastScanTime = 0

-- Last unit that was queried
local lastUnitQueried = nil

-----------------------------------------------------------------------
-- Init
--

local broker = LibStub("LibDataBroker-1.1")
local imba = broker:NewDataObject("Imba", {
	icon = "Interface\\AddOns\\Broker_CMA\\icon",
	type = "data source",
	text = "Imba",
	OnTooltipShow = function(tt)
		tt:AddLine("Broker: Imba!")
		if #missingBuffs > 0 then
			tt:AddLine("Missing buffs")
			tt:AddLine(table.concat(missingBuffs, ", "), 1, 1, 1, 1)
			tt:AddLine(" ")
		end
		if #missingDebuffs > 0 then
			tt:AddLine("Missing debuffs")
			tt:AddLine(table.concat(missingDebuffs, ", "), 1, 1, 1, 1)
			tt:AddLine(" ")
		end
		if next(watch) then
			tt:AddLine("Some units are out of range for talent scans. Totals and tooltip will be updated once they are in range.", 1, 0, 0, 1)
			tt:AddLine(" ")
		end
		tt:AddLine("If you don't know who can provide which of the missing buffs or debuffs, ask your group.", 0.2, 1, 0.2, 1)
	end,
})
addon:SetScript("OnEvent", function(self, event, ...)
	self[event](self, event, ...)
end)
addon:RegisterEvent("VARIABLES_LOADED")
addon:RegisterEvent("UNIT_PET")
addon:RegisterEvent("INSPECT_READY")
function addon:VARIABLES_LOADED() end

-----------------------------------------------------------------------
-- Display
--

local function updateText()
	if abilitiesCovered == 23 then
		imba.text = "All covered"
	elseif next(watch) then
		imba.text = ("|cffff0000!|r %d/23"):format(abilitiesCovered)
	else
		imba.text = ("%d/23"):format(abilitiesCovered)
	end
end

local bb = bit.band
local tmp = {}
local function updateTotals()
	wipe(missingBuffs)
	wipe(missingDebuffs)
	abilitiesCovered = 0

	for ability, B in pairs(abilityMap) do
		for unit, provides in pairs(providers) do
			if bb(provides, B) == B then
				--print(unit .. " provides " .. ability)
				tmp[ability] = true
			end
		end
	end
	for ability in pairs(abilityMap) do
		if not tmp[ability] then
			if buffs[ability] then missingBuffs[#missingBuffs + 1] = ability
			elseif debuffs[ability] then missingDebuffs[#missingDebuffs + 1] = ability
			end
		else
			abilitiesCovered = abilitiesCovered + 1
		end
	end
	table.sort(missingDebuffs)
	table.sort(missingBuffs)

	updateText()

	wipe(tmp)
end

-----------------------------------------------------------------------
-- Scanning
--

hooksecurefunc("NotifyInspect", function(unit)
	lastUnitQueried = UnitName(unit)
	lastScanTime = GetTime()
end)
local function queryUnit(unit)
	print("queryUnit: " .. unit)
	NotifyInspect(unit)
end

local function scanNext()
	if lastUnitQueried or not next(needsScanning) then return end
	local unit = next(needsScanning)
	if UnitExists(unit) and UnitIsVisible(unit) and UnitIsConnected(unit) and CheckInteractDistance(unit, 4) then
		if GetTime() > (lastScanTime + 1) then
			needsScanning[unit] = nil
			queryUnit(unit)
		else
			addon:Show()
		end
	else
		needsScanning[unit] = nil
		print("watch: " .. unit)
		if not next(watch) then
			watch[unit] = true
			updateText()
		else
			watch[unit] = true
		end
		-- I think we only need UNIT_AURA
		addon:RegisterEvent("UNIT_AURA")
		addon:RegisterEvent("UNIT_CONNECTION")
		addon:RegisterEvent("UNIT_NAME_UPDATE")
		scanNext()
	end
end

do
	local total = 0
	addon:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total > 1 then
			self:Hide()
			total = 0
			scanNext()
		end
	end)
	addon:Hide()
end

local function stopWatching()
	addon:UnregisterEvent("UNIT_AURA")
	addon:UnregisterEvent("UNIT_CONNECTION")
	addon:UnregisterEvent("UNIT_NAME_UPDATE")
	updateText()
end

-----------------------------------------------------------------------
-- Roster
--

local function scanRoster()
	for i = 1, GetNumRaidMembers() do
		local n = GetRaidRosterInfo(i)
		if n and not providers[n] and not watch[n] and lastUnitQueried ~= n then
			needsScanning[n] = true
		end
	end
	for u in pairs(providers) do
		if not UnitExists(u) then
			providers[u] = nil
		end
	end
	for unit in pairs(watch) do
		if providers[unit] or needsScanning[unit] or not UnitExists(unit) then
			watch[unit] = nil
		end
	end

	if not next(watch) then stopWatching() end
	if next(needsScanning) then
		scanNext()
	else
		updateTotals()
	end
end

addon:RegisterEvent("RAID_ROSTER_UPDATE")
function addon:RAID_ROSTER_UPDATE() scanRoster() end
addon.PARTY_MEMBERS_CHANGED = addon.RAID_ROSTER_UPDATE

-----------------------------------------------------------------------
-- Force updates
--

do
	local specChange = {}
	for specIndex, spellId in next, _G.TALENT_ACTIVATION_SPELLS do
		specChange[GetSpellInfo(spellId)] = specIndex
	end
	function addon:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell)
		if not specChange[spell] then return end
		local name = UnitName(unit)
		needsScanning[name] = true
		scanNext()
	end
end
addon:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")

function addon:UNIT_AURA(event, unit)
	local n = UnitName(unit)
	if not watch[n] then return end
	if UnitExists(unit) and UnitIsVisible(unit) and UnitIsConnected(unit) and CheckInteractDistance(unit, 4) then
		print("Scan for " .. n .. " found from " .. event)
		needsScanning[n] = true
		watch[n] = nil
		if not next(watch) then stopWatching() end
		scanNext()
	end
end
addon.UNIT_CONNECTION = addon.UNIT_AURA
addon.UNIT_NAME_UPDATE = addon.UNIT_AURA

-----------------------------------------------------------------------
-- Provider calculation
--

local function getUnitNameFromGUID(guid)
	for i = 1, GetNumRaidMembers() do
		local n = GetRaidRosterInfo(i)
		if n and UnitGUID(n) == guid then return n end
	end
end

function addon:INSPECT_READY(event, guid)
	if not guid then
		print("Got no GUID to INSPECT_READY, so aborting.")
		return
	end
	if not lastUnitQueried or guid ~= UnitGUID(lastUnitQueried) then
		-- Someone else is messing with the inspects.
		-- Lets see if we can find this GUID in our group,
		-- and if we can, then just continue and update
		-- the providers.
		lastUnitQueried = getUnitNameFromGUID(guid)
	end
	if not lastUnitQueried then
		scanNext()
		return
	end

	local group = GetActiveTalentGroup(true)
	local tree = 1
	for i = 1, 3 do
		local _,_,_,_, pointsSpent = GetTalentTabInfo(i, true, nil, group)
		if pointsSpent > 30 then
			tree = i + 1
			break
		end
	end
	local _, class = UnitClass(lastUnitQueried)
	if not class or not totalBits[class] or not totalBits[class][tree] then return end
	local bits = totalBits[class][tree]
	if providers[lastUnitQueried] ~= bits then
		providers[lastUnitQueried] = totalBits[class][tree]
		updateTotals()
	end
	lastUnitQueried = nil

	scanNext()
end

function addon:UNIT_PET(event, unit)
	local petId = unit.."-pet"
	local family = UnitCreatureFamily(petId)
	if not totalBits[family] then return end
	if providers[petId] ~= totalBits[family] then
		providers[petId] = totalBits[family]
		updateTotals()
	end
end

